#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import unittest
import time
import json
import threading

from .queue_client import QueueClient
from .queue_client import PredictException

logging.basicConfig(level=logging.ERROR,
                    format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
logger = logging.getLogger(__name__)


QueueEndpoint = '1828488879222746.cn-shanghai.pai-eas.aliyuncs.com'
QueueName = 'test_group.qservice'
SinkQueueName = 'test_group.qservice/sink'
QueueToken = ''

input_queue = None
sink_queue = None


class QueueClientTestCase(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super(QueueClientTestCase, self).__init__(*args, **kwargs)
        global input_queue
        global sink_queue
        if input_queue is None:
            input_queue = QueueClient(QueueEndpoint, QueueName)
            input_queue.set_token(QueueToken)
            input_queue.init()
        if sink_queue is None:
            sink_queue = QueueClient(QueueEndpoint, SinkQueueName)
            sink_queue.set_token(QueueToken)
            sink_queue.init()

        self.input_queue = input_queue
        self.sink_queue = sink_queue

    def truncate(self):
        attributes = self.input_queue.attributes()
        if 'stream.lastEntry' in attributes:
            self.input_queue.truncate(int(attributes['stream.lastEntry']) + 1)

        attributes = self.sink_queue.attributes()
        if 'stream.lastEntry' in attributes:
            self.sink_queue.truncate(int(attributes['stream.lastEntry']) + 1)

    def test_queue_get_by_request_id(self):
        self.truncate()

        index, request_id = self.sink_queue.put('abc')

        res_list = self.sink_queue.get(request_id=request_id)
        self.assertEqual(len(res_list), 1)
        self.assertEqual(res_list[0].data.decode('utf-8'), 'abc')

    def test_queue_get_by_index(self):
        index, request_id = self.sink_queue.put('abc')

        res_list = self.sink_queue.get(index=index)
        self.assertEqual(len(res_list), 1)
        self.assertEqual(res_list[0].data.decode('utf-8'), 'abc')

    def test_queue_truncate(self):
        latest_index = 0
        for x in range(10):
            index, request_id = self.sink_queue.put('abc')
            latest_index = index

        attributes = self.sink_queue.attributes()
        self.assertTrue(int(attributes['stream.length']) > 1)

        self.sink_queue.truncate(int(latest_index) + 1)

        attributes = self.sink_queue.attributes()

        self.assertEqual(int(attributes['stream.length']), 0)

    def test_watch_with_auto_commit(self):

        for x in range(10):
            index, request_id = self.sink_queue.put(str(x))

        i = 0
        watcher = self.sink_queue.watch(0, 5, auto_commit=True)
        for x in watcher.run():
            self.assertEqual(i, int(x.data.decode('utf-8')))
            i += 1
            if i == 10:
                break

        watcher.close()
        attributes = self.sink_queue.attributes()

        time.sleep(1)

        attributes = self.sink_queue.attributes()

        self.assertEqual(int(attributes['stream.length']), 0)

    def test_watch_with_manual_commit(self):

        for x in range(10):
            index, request_id = self.sink_queue.put(str(x))

        attributes = self.sink_queue.attributes()

        i = 0
        watcher = self.sink_queue.watch(0, 5, auto_commit=False)
        for x in watcher.run():
            self.assertEqual(i, int(x.data.decode('utf-8')))
            i += 1
            self.sink_queue.commit(x.index)
            if i == 10:
                break

        watcher.close()

        attributes = self.sink_queue.attributes()

        self.assertEqual(int(attributes['stream.length']), 0)

    def test_watch_with_inference_service_sync(self):
        self.truncate()

        for x in range(100):
            index, request_id = self.input_queue.put('[{}]')

        attributes = self.input_queue.attributes()

        i = 0
        watcher = self.sink_queue.watch(0, 5, auto_commit=False)
        for x in watcher.run():
            js = json.loads(x.data.decode('utf-8'))
            self.assertEqual(len(js[0].items()), 2)
            self.sink_queue.commit(x.index)
            i += 1
            if i == 100:
                break

        watcher.close()

        attributes = self.sink_queue.attributes()

        self.assertEqual(int(attributes['stream.length']), 0)

    def test_watch_async(self):
        self.truncate()

        count = 2000

        send_items = set()
        recv_items = set()

        def send_thread():
            self.sink_queue.set_timeout(30000)
            for x in range(count):
                index, request_id = self.sink_queue.put('[{}]')
                send_items.add(request_id)

        def watch_thread():
            watcher = self.sink_queue.watch(0, 5, auto_commit=True)
            i = 0
            for x in watcher.run():
                recv_items.add(x.tags['requestId'])
                i += 1
                if i == count:
                    break
            watcher.close()

        thread1 = threading.Thread(target=watch_thread)
        thread2 = threading.Thread(target=send_thread)

        thread1.start()
        thread2.start()

        thread1.join()
        thread2.join()

        self.assertEqual(send_items, recv_items)

    def test_watch_with_inference_service_async(self):
        self.truncate()

        count = 2000

        send_items = set()
        recv_items = set()

        def send_thread():
            self.input_queue.set_timeout(30000)
            for x in range(count):
                index, request_id = self.input_queue.put('[{}]')
                send_items.add(request_id)

        def watch_thread():
            watcher = self.sink_queue.watch(0, 5, auto_commit=True)
            i = 0
            for x in watcher.run():
                recv_items.add(x.tags['requestId'])
                i += 1
                if i == count:
                    break
            watcher.close()

        thread1 = threading.Thread(target=watch_thread)
        thread2 = threading.Thread(target=send_thread)

        thread1.start()
        thread2.start()

        thread1.join()
        thread2.join()

        self.assertEqual(send_items, recv_items)

    def test_watch_with_two_watcher(self):
        for x in range(1):
            index, request_id = self.sink_queue.put('[{}]')

        watcher = self.sink_queue.watch(0, 5, auto_commit=True)
        for x in watcher.run():
            break
        with self.assertRaises(PredictException):
            for x in watcher.run():
                break
        watcher.close()


if __name__ == '__main__':
    unittest.main()



# -*- encoding:utf-8 -*-
# Copyright (c) Alibaba, Inc. and its affiliates.
import argparse
import logging
import sys
import traceback

from eas_prediction.easyrec_request import EasyRecRequest
from eas_prediction.easyrec_predict_pb2 import PBFeature
from eas_prediction.easyrec_predict_pb2 import PBRequest

logging.basicConfig(
    level=logging.INFO, format='[%(asctime)s][%(levelname)s] %(message)s')

try:
  from eas_prediction import PredictClient  # TFRequest
except Exception:
  logging.error('eas_prediction is not installed: pip install eas-prediction')
  sys.exit(1)


def build_request(table_cols, table_data, item_ids=None):
  request_pb = PBRequest()
  assert isinstance(table_data, list)
  try:
    for col_id in range(len(table_cols)):
      cname, dtype = table_cols[col_id]
      value = table_data[col_id]
      feat = PBFeature()
      if value is None:
        continue
      if dtype == 'STRING':
        feat.string_feature = value
      elif dtype in ('FLOAT', 'DOUBLE'):
        feat.float_feature = float(value)
      elif dtype == 'BIGINT':
        feat.long_feature = int(value)
      elif dtype == 'INT':
        feat.int_feature = int(value)

      request_pb.user_features[cname].CopyFrom(feat)
  except Exception:
    traceback.print_exc()
    sys.exit()
  request_pb.item_ids.extend(item_ids)
  return request_pb


def parse_table_schema(create_table_sql):
  create_table_sql = create_table_sql.lower()
  spos = create_table_sql.index('(')
  epos = create_table_sql[spos + 1:].index(')') + spos + 1
  cols = create_table_sql[(spos + 1):epos]
  cols = [x.strip().lower() for x in cols.split(',')]
  col_info_arr = []
  for col in cols:
    col = [k for k in col.split() if k != '']
    assert len(col) == 2
    col[1] = col[1].upper()
    col_info_arr.append(col)
  return col_info_arr


def send_request(req_pb, client, debug_level=0):
  req = EasyRecRequest()
  req.add_feed(req_pb, debug_level)
  tmp = client.predict(req)
  return tmp


if __name__ == '__main__':
  parser = argparse.ArgumentParser()
  parser.add_argument(
      '--endpoint',
      type=str,
      default=None,
      help='eas endpoint, such as 12345.cn-beijing.pai-eas.aliyuncs.com')
  parser.add_argument(
      '--service_name', type=str, default=None, help='eas service name')
  parser.add_argument(
      '--token', type=str, default=None, help='eas service token')
  parser.add_argument(
      '--table_schema',
      type=str,
      default=None,
      help='user feature table schema path, example: create table(user_id string, age bigint);')
  parser.add_argument(
      '--table_data',
      type=str,
      default=None,
      help='user feature table data path, content are user_feature values separated by ; , the number and types of feature values should match the definitions in table_schema')
  parser.add_argument('--item_lst', type=str, default=None, help='item list file path, content are item_ids, each in one line')

  args, _ = parser.parse_known_args()

  if args.endpoint is None:
    logging.error('--endpoint is not set')
    sys.exit(1)
  if args.service_name is None:
    logging.error('--service_name is not set')
    sys.exit(1)
  if args.token is None:
    logging.error('--token is not set')
    sys.exit(1)
  if args.table_schema is None:
    logging.error('--table_schema is not set')
    sys.exit(1)
  if args.table_data is None:
    logging.error('--table_data is not set')
    sys.exit(1)
  if args.item_lst is None:
    logging.error('--item_lst is not set')
    sys.exit(1)

  client = PredictClient(args.endpoint, args.service_name)
  client.set_token(args.token)
  client.init()

  with open(args.table_schema, 'r') as fin:
    create_table_sql = fin.read().strip()
  table_cols = parse_table_schema(create_table_sql)

  table_data = []
  with open(args.table_data, 'r') as fin:
    for line_str in fin:
      table_data.append(line_str.strip())

  assert len(table_cols) == len(table_data), 'user features[%s] not match with schema[%s]' % (table_data, table_cols)

  items = []
  with open(args.item_lst, 'r') as fin:
    for line_str in fin:
      items.append(line_str.strip())
 

  req = build_request(table_cols, table_data, item_ids=items)
  resp = send_request(req, client)
  logging.info(resp)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pytorch_predict.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='pytorch_predict.proto',
  package='pytorch.eas',
  syntax='proto3',
  serialized_options=_b('\370\001\001'),
  serialized_pb=_b('\n\x15pytorch_predict.proto\x12\x0bpytorch.eas\"\x1d\n\nArrayShape\x12\x0f\n\x03\x64im\x18\x01 \x03(\x03\x42\x02\x10\x01\"\xd4\x01\n\nArrayProto\x12)\n\x05\x64type\x18\x01 \x01(\x0e\x32\x1a.pytorch.eas.ArrayDataType\x12,\n\x0b\x61rray_shape\x18\x02 \x01(\x0b\x32\x17.pytorch.eas.ArrayShape\x12\x15\n\tfloat_val\x18\x03 \x03(\x02\x42\x02\x10\x01\x12\x16\n\ndouble_val\x18\x04 \x03(\x01\x42\x02\x10\x01\x12\x13\n\x07int_val\x18\x05 \x03(\x05\x42\x02\x10\x01\x12\x12\n\nstring_val\x18\x06 \x03(\x0c\x12\x15\n\tint64_val\x18\x07 \x03(\x03\x42\x02\x10\x01\"P\n\x0ePredictRequest\x12\'\n\x06inputs\x18\x01 \x03(\x0b\x32\x17.pytorch.eas.ArrayProto\x12\x15\n\routput_filter\x18\x02 \x03(\x05\";\n\x0fPredictResponse\x12(\n\x07outputs\x18\x01 \x03(\x0b\x32\x17.pytorch.eas.ArrayProto*\xdb\x02\n\rArrayDataType\x12\x0e\n\nDT_INVALID\x10\x00\x12\x0c\n\x08\x44T_FLOAT\x10\x01\x12\r\n\tDT_DOUBLE\x10\x02\x12\x0c\n\x08\x44T_INT32\x10\x03\x12\x0c\n\x08\x44T_UINT8\x10\x04\x12\x0c\n\x08\x44T_INT16\x10\x05\x12\x0b\n\x07\x44T_INT8\x10\x06\x12\r\n\tDT_STRING\x10\x07\x12\x10\n\x0c\x44T_COMPLEX64\x10\x08\x12\x0c\n\x08\x44T_INT64\x10\t\x12\x0b\n\x07\x44T_BOOL\x10\n\x12\x0c\n\x08\x44T_QINT8\x10\x0b\x12\r\n\tDT_QUINT8\x10\x0c\x12\r\n\tDT_QINT32\x10\r\x12\x0f\n\x0b\x44T_BFLOAT16\x10\x0e\x12\r\n\tDT_QINT16\x10\x0f\x12\x0e\n\nDT_QUINT16\x10\x10\x12\r\n\tDT_UINT16\x10\x11\x12\x11\n\rDT_COMPLEX128\x10\x12\x12\x0b\n\x07\x44T_HALF\x10\x13\x12\x0f\n\x0b\x44T_RESOURCE\x10\x14\x12\x0e\n\nDT_VARIANT\x10\x15\x42\x03\xf8\x01\x01\x62\x06proto3')
)

_ARRAYDATATYPE = _descriptor.EnumDescriptor(
  name='ArrayDataType',
  full_name='pytorch.eas.ArrayDataType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='DT_INVALID', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DT_FLOAT', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DT_DOUBLE', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DT_INT32', index=3, number=3,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DT_UINT8', index=4, number=4,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DT_INT16', index=5, number=5,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DT_INT8', index=6, number=6,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DT_STRING', index=7, number=7,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DT_COMPLEX64', index=8, number=8,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DT_INT64', index=9, number=9,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DT_BOOL', index=10, number=10,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DT_QINT8', index=11, number=11,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DT_QUINT8', index=12, number=12,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DT_QINT32', index=13, number=13,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DT_BFLOAT16', index=14, number=14,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DT_QINT16', index=15, number=15,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DT_QUINT16', index=16, number=16,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DT_UINT16', index=17, number=17,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DT_COMPLEX128', index=18, number=18,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DT_HALF', index=19, number=19,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DT_RESOURCE', index=20, number=20,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DT_VARIANT', index=21, number=21,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=428,
  serialized_end=775,
)
_sym_db.RegisterEnumDescriptor(_ARRAYDATATYPE)

ArrayDataType = enum_type_wrapper.EnumTypeWrapper(_ARRAYDATATYPE)
DT_INVALID = 0
DT_FLOAT = 1
DT_DOUBLE = 2
DT_INT32 = 3
DT_UINT8 = 4
DT_INT16 = 5
DT_INT8 = 6
DT_STRING = 7
DT_COMPLEX64 = 8
DT_INT64 = 9
DT_BOOL = 10
DT_QINT8 = 11
DT_QUINT8 = 12
DT_QINT32 = 13
DT_BFLOAT16 = 14
DT_QINT16 = 15
DT_QUINT16 = 16
DT_UINT16 = 17
DT_COMPLEX128 = 18
DT_HALF = 19
DT_RESOURCE = 20
DT_VARIANT = 21



_ARRAYSHAPE = _descriptor.Descriptor(
  name='ArrayShape',
  full_name='pytorch.eas.ArrayShape',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='dim', full_name='pytorch.eas.ArrayShape.dim', index=0,
      number=1, type=3, cpp_type=2, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=_b('\020\001'), file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=38,
  serialized_end=67,
)


_ARRAYPROTO = _descriptor.Descriptor(
  name='ArrayProto',
  full_name='pytorch.eas.ArrayProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='dtype', full_name='pytorch.eas.ArrayProto.dtype', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='array_shape', full_name='pytorch.eas.ArrayProto.array_shape', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='float_val', full_name='pytorch.eas.ArrayProto.float_val', index=2,
      number=3, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=_b('\020\001'), file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='double_val', full_name='pytorch.eas.ArrayProto.double_val', index=3,
      number=4, type=1, cpp_type=5, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=_b('\020\001'), file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='int_val', full_name='pytorch.eas.ArrayProto.int_val', index=4,
      number=5, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=_b('\020\001'), file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='string_val', full_name='pytorch.eas.ArrayProto.string_val', index=5,
      number=6, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='int64_val', full_name='pytorch.eas.ArrayProto.int64_val', index=6,
      number=7, type=3, cpp_type=2, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=_b('\020\001'), file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=70,
  serialized_end=282,
)


_PREDICTREQUEST = _descriptor.Descriptor(
  name='PredictRequest',
  full_name='pytorch.eas.PredictRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='inputs', full_name='pytorch.eas.PredictRequest.inputs', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='output_filter', full_name='pytorch.eas.PredictRequest.output_filter', index=1,
      number=2, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=284,
  serialized_end=364,
)


_PREDICTRESPONSE = _descriptor.Descriptor(
  name='PredictResponse',
  full_name='pytorch.eas.PredictResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='outputs', full_name='pytorch.eas.PredictResponse.outputs', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=366,
  serialized_end=425,
)

_ARRAYPROTO.fields_by_name['dtype'].enum_type = _ARRAYDATATYPE
_ARRAYPROTO.fields_by_name['array_shape'].message_type = _ARRAYSHAPE
_PREDICTREQUEST.fields_by_name['inputs'].message_type = _ARRAYPROTO
_PREDICTRESPONSE.fields_by_name['outputs'].message_type = _ARRAYPROTO
DESCRIPTOR.message_types_by_name['ArrayShape'] = _ARRAYSHAPE
DESCRIPTOR.message_types_by_name['ArrayProto'] = _ARRAYPROTO
DESCRIPTOR.message_types_by_name['PredictRequest'] = _PREDICTREQUEST
DESCRIPTOR.message_types_by_name['PredictResponse'] = _PREDICTRESPONSE
DESCRIPTOR.enum_types_by_name['ArrayDataType'] = _ARRAYDATATYPE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ArrayShape = _reflection.GeneratedProtocolMessageType('ArrayShape', (_message.Message,), {
  'DESCRIPTOR' : _ARRAYSHAPE,
  '__module__' : 'pytorch_predict_pb2'
  # @@protoc_insertion_point(class_scope:pytorch.eas.ArrayShape)
  })
_sym_db.RegisterMessage(ArrayShape)

ArrayProto = _reflection.GeneratedProtocolMessageType('ArrayProto', (_message.Message,), {
  'DESCRIPTOR' : _ARRAYPROTO,
  '__module__' : 'pytorch_predict_pb2'
  # @@protoc_insertion_point(class_scope:pytorch.eas.ArrayProto)
  })
_sym_db.RegisterMessage(ArrayProto)

PredictRequest = _reflection.GeneratedProtocolMessageType('PredictRequest', (_message.Message,), {
  'DESCRIPTOR' : _PREDICTREQUEST,
  '__module__' : 'pytorch_predict_pb2'
  # @@protoc_insertion_point(class_scope:pytorch.eas.PredictRequest)
  })
_sym_db.RegisterMessage(PredictRequest)

PredictResponse = _reflection.GeneratedProtocolMessageType('PredictResponse', (_message.Message,), {
  'DESCRIPTOR' : _PREDICTRESPONSE,
  '__module__' : 'pytorch_predict_pb2'
  # @@protoc_insertion_point(class_scope:pytorch.eas.PredictResponse)
  })
_sym_db.RegisterMessage(PredictResponse)


DESCRIPTOR._options = None
_ARRAYSHAPE.fields_by_name['dim']._options = None
_ARRAYPROTO.fields_by_name['float_val']._options = None
_ARRAYPROTO.fields_by_name['double_val']._options = None
_ARRAYPROTO.fields_by_name['int_val']._options = None
_ARRAYPROTO.fields_by_name['int64_val']._options = None
# @@protoc_insertion_point(module_scope)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .request import Request
from .request import Response
from . import blade_pb2
from functools import reduce 
import operator


class TFRequest(Request):
    """
    Request for tensorflow services whose input data is in format of protobuf,
    privide methods to generate the required protobuf object, and serialze it to string
    """
    DT_FLOAT = blade_pb2.DT_FLOAT
    DT_INT32 = blade_pb2.DT_INT32
    DT_INT64 = blade_pb2.DT_INT64
    DT_STRING = blade_pb2.DT_STRING
    DT_UNKNOWN = blade_pb2.DT_UNKNOWN

    def __init__(self, signature_name=None):
        self.request_data = blade_pb2.Request()
        self.set_signature_name(signature_name)

    def __str__(self):
        return self.request_data

    def set_signature_name(self, signature_name):
        """
        Set the signature name of the model
        :param signature_name: signature name of the tensorflow model
        """
        self.tf_signature_name = signature_name
        self.request_data.tf_signature_name = self.tf_signature_name

    def volume(self, shape):
        return reduce(operator.mul, shape)

    def add_feed(self, input_name, shape, content_type, content):
        """
        Add input data for the request, a tensorflow model may have many inputs with different
        data types, this methods set data for one of the input with the specified name 'input_name'
        :param input_name: name of the input to be set
        :param shape: shape of the input tensor in format of array, such as [1,784]
        :param content_type: type of the input tensor (predefined data type, such as TFRequest.DT_FLOAT)
        :param content: data content of the input tensor, the content could be a single value (e.g. 3.14), or a expanded one-dimension array (e.g. [1,2,3,4,5])
        (NOTE: blade.request needs param 'batchsize', while to keep consistent with EAS default TFRequest api, we removed this argument, by deduce the 'batchsize' from shape's first dimention.)
        """
        if (not input_name):
            raise AssertionError('"input name" can not be emtpy')
        if (not isinstance(shape, list)):
            raise AssertionError('Invalid type for "shape": {}'.format(type(shape)))
        arr = self.request_data.inputs.add()
        arr.shape.dim.extend(shape)
        arr.name_in_dl_model = input_name
        arr.blade_blob_name = input_name # HERE: 'blade blob name' is identical to 'name in DL model'.
        arr.batchsize = int(shape[0]) # HERE: deduce batchsize from shape's first dimension
        if (content_type == blade_pb2.DT_FLOAT):
            if (isinstance(content, float)):
                if (int(self.volume(shape)) == 1):
                    arr.float_val.append(content)
            elif (isinstance(content, list)):
                if (int(self.volume(shape)) == int(len(content))):
                    arr.float_val.extend(content)
                else:
                    raise AssertionError('Shape volume {} mis-match content size {}'.format( \
                            self.volume(shape), len(content)))
            else:
                raise AssertionError('Expect float or list of float type')
        elif (content_type == blade_pb2.DT_INT32):
            if (isinstance(content, int)):
                if (int(self.volume(shape)) == 1):
                    arr.int_val.append(content)
                else:
                    raise AssertionError('Shape volume {} mis-match content size {}'.format( \
                            self.volume(shape), 1))
            elif (isinstance(content, list)):
                if (int(self.volume(shape)) == int(len(content))):
                    arr.int_val.extend(content)
                else:
                    raise AssertionError('Shape volume {} mis-match content size {}'.format( \
                            self.volume(shape), len(content)))
            else:
                raise AssertionError('Expect int32 or list of int32 type')
        elif (content_type == blade_pb2.DT_INT64):
            if (isinstance(content, int)):
                if (int(self.volume(shape)) == 1):
                    arr.in64_val.append(content)
                else:
                    raise AssertionError('Shape volume {} mis-match content size {}'.format( \
                            self.volume(shape), 1))
            elif (isinstance(content, list)):
                if (int(self.volume(shape)) == int(len(content))):
                    arr.int64_val.extend(content)
                else:
                    raise AssertionError('Shape volume {} mis-match content size {}'.format( \
                            self.volume(shape), len(content)))
            else:
                raise AssertionError('Expect int64 or list of int64 type')
        elif (content_type == blade_pb2.DT_STRING):
            if (isinstance(content, str)):
                if (int(self.volume(shape)) == 1):
                    arr.string_val.append(content)
                else:
                    raise AssertionError('Shape volume {} mis-match content size {}'.format( \
                            self.volume(shape), 1))
            elif (isinstance(content, list)):
                if (int(self.volume(shape)) == int(len(content))):
                    arr.string_val.extend(content)
                else:
                    raise AssertionError('Shape volume {} mis-match content size {}'.format( \
                            self.volume(shape), len(content)))
            else:
                raise AssertionError('Expect str or list of str type')
        else:
            raise AssertionError('Unknown data_type: "{}"'.format(data_type))
        return

    def add_fetch(self, output_name):
        """
        Add output node name for the request to get, if not specified, then all the known outputs are fetched,
        but for frozen models, the output name must be specified, or else the service would throw exception like:
        'Must specify at least one target to fetch or execute.'
        :param output_name: name of the output node to fetch
        (NOTE: blade.request output_info has field output_type, while to keep consistent with EAS default TFRequest api, here we chooce to ignore it.)
        """
        if (not output_name):
            raise AssertionError('"input name" can not be emtpy')
        out_info = self.request_data.output_info.add()
        out_info.blade_blob_name = output_name
        out_info.name_in_dl_model = output_name
        out_info.data_type = TFRequest.DT_UNKNOWN
        return

    def to_string(self):
        """
        Serialize the request to string for transmission
        :return: the request data in format of string
        """
        return self.request_data.SerializeToString()

    def parse_response(self, response_data):
        """
        Parse the given response data in string format to the related TFResponse object
        :param response_data: the service response data in string format
        :return: the TFResponse object related the request
        """
        return TFResponse(response_data)


class TFResponse(Response):
    """
    Deserialize the response data to a structured object for users to read
    """

    def __init__(self, response_data=None):
        self.response = blade_pb2.Response()
        self.response.ParseFromString(response_data)

    def __str__(self):
        return str(self.response)

    def get_tensor_shape(self, output_name):
        """
        Get the shape of a specified output tensor
        :param output_name: name of the output tensor
        :return: the shape in format of list
        """
        for out in self.response.outputs:
            if (str(out.name_in_dl_model).strip() == str(output_name).strip()):
                return list(out.shape.dim)
        raise AssertionError('Unknown output name: "{}"'.format(output_name))

    def get_values(self, output_name):
        """
        Get the value of a specified output tensor
        :param output_name: name of the output tensor
        :return: the content of the output tensor
        (NOTE: blade.request output_info has field output_type, while to keep consistent with EAS default TFResponse api, here we choose to ignore it.)
        """
        for out in self.response.outputs:
            if (str(out.name_in_dl_model).strip() == str(output_name).strip()):
                if (len(out.float_val) > 0):
                    return out.float_val
                if (len(out.int_val) > 0):
                    return out.int_val
                if (len(out.int64_val) > 0):
                    return out.int64_val
                if (len(out.string_val) > 0):
                    return out.string_val
        raise AssertionError('Unknown output name: "{}"'.format(output_name))



#########
#  Author: Marcin Matlaszek <mmatlaszek@gmail.com>
#  Description: Implementation of WeightedRoundRobin algorithm.
########

import sys
import math
import fractions
import functools


class WRRScheduler:
    cw = 0
    i = -1
    data_set = []
    max_s = None
    gcd_s = None
    len_s = None

    def __init__(self, s=None):
        self._init_dataset(s)

    def _init_dataset(self, s):
        self.data_set = s
        if s is None or len(s) == 0:
            return
        self.max_s = max(s, key=lambda x: x[1])[1]
        if sys.version_info[0] < 3:
            self.gcd_s = functools.reduce(fractions.gcd, [weight for data, weight in s])
        else:
            self.gcd_s = functools.reduce(math.gcd, [weight for data, weight in s])
        self.len_s = len(s)

    def schedule(self):
        if self.data_set is None or len(self.data_set) == 0:
            return None
        while True:
            self.i = (self.i + 1) % self.len_s
            if self.i == 0:
                self.cw = self.cw - self.gcd_s
                if self.cw <= 0:
                    self.cw = self.max_s
                    if self.cw == 0:
                        return None
            if self.data_set[self.i][1] >= self.cw:
                return self.data_set[self.i]

    def set_data(self, s):
        self.reset()
        self._init_dataset(s)

    def reset(self):
        self.cw = 0
        self.i = -1
        self.data_set = []
        self.max_s = None
        self.gcd_s = None
        self.len_s = None

    def get_next(self, n=1):
        if n > 1:
            return [self.schedule() for i in range(0, n)]
        return self.schedule()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .request import Request
from .request import Response


class StringRequest(Request):
    """
    Request contains raw string data
    """
    def __init__(self, request_data=None):
        self.request_data = '' if request_data is None else request_data

    def __str__(self):
        return self.request_data

    def to_string(self):
        """
        Get the request data in string format
        :return: the request data in format of string
        """
        return self.request_data

    def parse_response(self, response_data):
        """
        Parse the given response data in string format to the related StringResponse object
        :param response_data: the service response data in string format
        :return: the StringResponse object related the request
        """
        return StringResponse(response_data)


class StringResponse(Response):
    """
    Deserialize the response data to a structured object for users to read
    """
    def __init__(self, response_data=None):
        self.response_data = '' if response_data is None else response_data

    def __str__(self):
        return str(self.response_data)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .request import Request
from .request import Response
from . import pytorch_predict_pb2 as pt_pb


class TorchRequest(Request):
    """
    Request for pytorch services whose input data is in format of protobuf,
    privide methods to generate the required protobuf object, and serialze it to string
    """
    DT_FLOAT = pt_pb.DT_FLOAT
    DT_DOUBLE = pt_pb.DT_DOUBLE
    DT_UINT8 = pt_pb.DT_UINT8
    DT_INT8 = pt_pb.DT_INT8
    DT_INT16 = pt_pb.DT_INT16
    DT_INT32 = pt_pb.DT_INT32
    DT_INT64 = pt_pb.DT_INT64

    def __init__(self):
        self.request_data = pt_pb.PredictRequest()
        # self.index = index

    def __str__(self):
        return self.request_data

    def add_feed(self, index, shape, content_type, content):
        """
        Add input data for the request, a pytorch model may have many inputs with different
        data types, this methods set data for one of the input with the specified 'index'
        :param index: name of the input to be set
        :param shape: shape of the input tensor in format of array, such as [1, 3, 224, 224]
        :param content_type: type of the input tensor, can be one of the predefined data type, such as TorchRequest.DT_FLOAT
        :param content: data content of the input tensor, which is expanded to one-dimension array, such as [1,2,3,4,5]
        """
        # self.request_data.index = self.index
        while len(self.request_data.inputs) < index + 1:
            self.request_data.inputs.add()
        self.request_data.inputs[index].dtype = content_type
        self.request_data.inputs[index].array_shape.dim.extend(shape)
        if content_type == pt_pb.DT_FLOAT:
            self.request_data.inputs[index].float_val.extend(content)
        elif content_type == pt_pb.DT_DOUBLE:
            self.request_data.inputs[index].double_val.extend(content)
        elif content_type == pt_pb.DT_INT8 or content_type == pt_pb.DT_INT16 or \
                content_type == pt_pb.DT_INT32 or content_type == pt_pb.DT_UINT8:
            self.request_data.inputs[index].int_val.extend(content)
        elif content_type == pt_pb.DT_INT64:
            self.request_data.inputs[index].int64_val.extend(content)

    def add_fetch(self, output_index):
        """
        Add output index for the request to get, if not specified, then all the known outputs are fetched
        :param index: index of the output node to fetch
        """
        self.request_data.output_filter.append(output_index)

    def to_string(self):
        """
        Serialize the request to string for transmission
        :return: the request data in format of string
        """
        return self.request_data.SerializeToString()

    def parse_response(self, response_data):
        """
        Parse the given response data in string format to the related TorchResponse object
        :param response_data: the service response data in string format
        :return: the TorchResponse object related the request
        """
        return TorchResponse(response_data)


class TorchResponse(Response):
    """
    Deserialize the response data to a structured object for users to read
    """

    def __init__(self, response_data=None):
        self.response = pt_pb.PredictResponse()
        self.response.ParseFromString(response_data)

    def __str__(self):
        return str(self.response)

    def get_tensor_shape(self, output_index):
        """
        Get the shape of a specified output tensor
        :param output_index: name of the output tensor
        :return: the shape in format of list
        """
        return list(self.response.outputs[output_index].array_shape.dim)

    def get_values(self, output_index):
        """
        Get the value of a specified output tensor
        :param output_index: name of the output tensor
        :return: the content of the output tensor
        """
        output = self.response.outputs[output_index]
        if output.dtype == TorchRequest.DT_FLOAT:
            return output.float_val
        elif output.dtype == TorchRequest.DT_INT8 or output.dtype == TorchRequest.DT_INT16 or \
                output.dtype == TorchRequest.DT_INT32 or output.dtype == TorchRequest.DT_UINT8:
            return output.int_val
        elif output.dtype == TorchRequest.DT_INT64:
            return output.int64_val
        elif output.dtype == TorchRequest.DT_DOUBLE:
            return output.double_val



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: hie_processor_data.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='hie_processor_data.proto',
  package='hie.processor',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=b'\n\x18hie_processor_data.proto\x12\rhie.processor\"\xe8\x01\n\x0bTensorProto\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0c\n\x04\x64ims\x18\x02 \x03(\x03\x12\x11\n\tdata_type\x18\x03 \x01(\x05\x12\x10\n\x08raw_data\x18\x05 \x01(\x0c\"\x97\x01\n\x08\x44\x61taType\x12\r\n\tUNDEFINED\x10\x00\x12\t\n\x05\x46LOAT\x10\x01\x12\t\n\x05UINT8\x10\x02\x12\x08\n\x04INT8\x10\x03\x12\n\n\x06UINT16\x10\x04\x12\t\n\x05INT16\x10\x05\x12\t\n\x05INT32\x10\x06\x12\t\n\x05INT64\x10\x07\x12\x0b\n\x07\x46LOAT16\x10\n\x12\n\n\x06\x44OUBLE\x10\x0b\x12\n\n\x06UINT32\x10\x0c\x12\n\n\x06UINT64\x10\r\"8\n\tDataProto\x12+\n\x07tensors\x18\x01 \x03(\x0b\x32\x1a.hie.processor.TensorProtob\x06proto3'
)



_TENSORPROTO_DATATYPE = _descriptor.EnumDescriptor(
  name='DataType',
  full_name='hie.processor.TensorProto.DataType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNDEFINED', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FLOAT', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='UINT8', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='INT8', index=3, number=3,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='UINT16', index=4, number=4,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='INT16', index=5, number=5,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='INT32', index=6, number=6,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='INT64', index=7, number=7,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FLOAT16', index=8, number=10,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DOUBLE', index=9, number=11,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='UINT32', index=10, number=12,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='UINT64', index=11, number=13,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=125,
  serialized_end=276,
)
_sym_db.RegisterEnumDescriptor(_TENSORPROTO_DATATYPE)


_TENSORPROTO = _descriptor.Descriptor(
  name='TensorProto',
  full_name='hie.processor.TensorProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='hie.processor.TensorProto.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='dims', full_name='hie.processor.TensorProto.dims', index=1,
      number=2, type=3, cpp_type=2, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='data_type', full_name='hie.processor.TensorProto.data_type', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='raw_data', full_name='hie.processor.TensorProto.raw_data', index=3,
      number=5, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _TENSORPROTO_DATATYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=44,
  serialized_end=276,
)


_DATAPROTO = _descriptor.Descriptor(
  name='DataProto',
  full_name='hie.processor.DataProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='tensors', full_name='hie.processor.DataProto.tensors', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=278,
  serialized_end=334,
)

_TENSORPROTO_DATATYPE.containing_type = _TENSORPROTO
_DATAPROTO.fields_by_name['tensors'].message_type = _TENSORPROTO
DESCRIPTOR.message_types_by_name['TensorProto'] = _TENSORPROTO
DESCRIPTOR.message_types_by_name['DataProto'] = _DATAPROTO
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

TensorProto = _reflection.GeneratedProtocolMessageType('TensorProto', (_message.Message,), {
  'DESCRIPTOR' : _TENSORPROTO,
  '__module__' : 'hie_processor_data_pb2'
  # @@protoc_insertion_point(class_scope:hie.processor.TensorProto)
  })
_sym_db.RegisterMessage(TensorProto)

DataProto = _reflection.GeneratedProtocolMessageType('DataProto', (_message.Message,), {
  'DESCRIPTOR' : _DATAPROTO,
  '__module__' : 'hie_processor_data_pb2'
  # @@protoc_insertion_point(class_scope:hie.processor.DataProto)
  })
_sym_db.RegisterMessage(DataProto)


# @@protoc_insertion_point(module_scope)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .endpoint import Endpoint
from .exception import PredictException
import os


class GatewayEndpoint(Endpoint):
    """
    Endpoint for the gateway access which provide a static http address for
    service accessing, there's no need for client to implement round robin,
    every time a request is invoked, just return the static http address.
    """
    def __init__(self, domain, service_name, logger):
        super(GatewayEndpoint, self).__init__(logger)
        if domain is None or len(domain) == 0:
            namespace = os.getenv('NAMESPACE')
            if namespace is None or len(namespace) == 0:
                raise PredictException(500, '\'endpoint\' must be set when running outside eas service')
            service_name = service_name.replace('_', '-')
            domain = '%s.%s:8080' % (service_name, namespace)

        if not domain.startswith('http://'):
            domain = 'http://' + domain
        if domain.endswith('/'):
            domain = domain[:len(domain)-1]
        self.domain = domain
        self.logger = logger

    def set_endpoints(self, endpoints):
        self.logger.debug('sync nothing for gateway endpoint')

    def get_size(self):
        return 1

    def get(self):
        return self.domain




# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: queueservice.proto

import sys

_b = sys.version_info[0] < 3 and (lambda x: x) or (lambda x: x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()

DESCRIPTOR = _descriptor.FileDescriptor(
  name='queueservice.proto',
  package='streaming',
  syntax='proto3',
  serialized_options=_b('Z\023pkg/streaming/types'),
  serialized_pb=_b(
    '\n\x12queueservice.proto\x12\tstreaming\"\x9e\x01\n\x0e\x44\x61taFrameProto\x12\r\n\x05index\x18\x01 \x01(\x04\x12\x31\n\x04tags\x18\x02 \x03(\x0b\x32#.streaming.DataFrameProto.TagsEntry\x12\x0c\n\x04\x64\x61ta\x18\x03 \x01(\x0c\x12\x0f\n\x07message\x18\x04 \x01(\t\x1a+\n\tTagsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\">\n\x12\x44\x61taFrameListProto\x12(\n\x05index\x18\x01 \x03(\x0b\x32\x19.streaming.DataFrameProto\"\x84\x01\n\x0f\x41ttributesProto\x12>\n\nattributes\x18\x01 \x03(\x0b\x32*.streaming.AttributesProto.AttributesEntry\x1a\x31\n\x0f\x41ttributesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x42\x15Z\x13pkg/streaming/typesb\x06proto3')
)

_DATAFRAMEPROTO_TAGSENTRY = _descriptor.Descriptor(
  name='TagsEntry',
  full_name='streaming.DataFrameProto.TagsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='streaming.DataFrameProto.TagsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='streaming.DataFrameProto.TagsEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=_b('8\001'),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=149,
  serialized_end=192,
)

_DATAFRAMEPROTO = _descriptor.Descriptor(
  name='DataFrameProto',
  full_name='streaming.DataFrameProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='index', full_name='streaming.DataFrameProto.index', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='tags', full_name='streaming.DataFrameProto.tags', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='data', full_name='streaming.DataFrameProto.data', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='message', full_name='streaming.DataFrameProto.message', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_DATAFRAMEPROTO_TAGSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=34,
  serialized_end=192,
)

_DATAFRAMELISTPROTO = _descriptor.Descriptor(
  name='DataFrameListProto',
  full_name='streaming.DataFrameListProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='index', full_name='streaming.DataFrameListProto.index', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=194,
  serialized_end=256,
)

_ATTRIBUTESPROTO_ATTRIBUTESENTRY = _descriptor.Descriptor(
  name='AttributesEntry',
  full_name='streaming.AttributesProto.AttributesEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='streaming.AttributesProto.AttributesEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='streaming.AttributesProto.AttributesEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=_b('8\001'),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=342,
  serialized_end=391,
)

_ATTRIBUTESPROTO = _descriptor.Descriptor(
  name='AttributesProto',
  full_name='streaming.AttributesProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='attributes', full_name='streaming.AttributesProto.attributes', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_ATTRIBUTESPROTO_ATTRIBUTESENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=259,
  serialized_end=391,
)

_DATAFRAMEPROTO_TAGSENTRY.containing_type = _DATAFRAMEPROTO
_DATAFRAMEPROTO.fields_by_name['tags'].message_type = _DATAFRAMEPROTO_TAGSENTRY
_DATAFRAMELISTPROTO.fields_by_name['index'].message_type = _DATAFRAMEPROTO
_ATTRIBUTESPROTO_ATTRIBUTESENTRY.containing_type = _ATTRIBUTESPROTO
_ATTRIBUTESPROTO.fields_by_name['attributes'].message_type = _ATTRIBUTESPROTO_ATTRIBUTESENTRY
DESCRIPTOR.message_types_by_name['DataFrameProto'] = _DATAFRAMEPROTO
DESCRIPTOR.message_types_by_name['DataFrameListProto'] = _DATAFRAMELISTPROTO
DESCRIPTOR.message_types_by_name['AttributesProto'] = _ATTRIBUTESPROTO
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

DataFrameProto = _reflection.GeneratedProtocolMessageType('DataFrameProto', (_message.Message,), dict(

  TagsEntry=_reflection.GeneratedProtocolMessageType('TagsEntry', (_message.Message,), dict(
    DESCRIPTOR=_DATAFRAMEPROTO_TAGSENTRY,
    __module__='queueservice_pb2'
    # @@protoc_insertion_point(class_scope:streaming.DataFrameProto.TagsEntry)
  ))
  ,
  DESCRIPTOR=_DATAFRAMEPROTO,
  __module__='queueservice_pb2'
  # @@protoc_insertion_point(class_scope:streaming.DataFrameProto)
))
_sym_db.RegisterMessage(DataFrameProto)
_sym_db.RegisterMessage(DataFrameProto.TagsEntry)

DataFrameListProto = _reflection.GeneratedProtocolMessageType('DataFrameListProto', (_message.Message,), dict(
  DESCRIPTOR=_DATAFRAMELISTPROTO,
  __module__='queueservice_pb2'
  # @@protoc_insertion_point(class_scope:streaming.DataFrameListProto)
))
_sym_db.RegisterMessage(DataFrameListProto)

AttributesProto = _reflection.GeneratedProtocolMessageType('AttributesProto', (_message.Message,), dict(

  AttributesEntry=_reflection.GeneratedProtocolMessageType('AttributesEntry', (_message.Message,), dict(
    DESCRIPTOR=_ATTRIBUTESPROTO_ATTRIBUTESENTRY,
    __module__='queueservice_pb2'
    # @@protoc_insertion_point(class_scope:streaming.AttributesProto.AttributesEntry)
  ))
  ,
  DESCRIPTOR=_ATTRIBUTESPROTO,
  __module__='queueservice_pb2'
  # @@protoc_insertion_point(class_scope:streaming.AttributesProto)
))
_sym_db.RegisterMessage(AttributesProto)
_sym_db.RegisterMessage(AttributesProto.AttributesEntry)

DESCRIPTOR._options = None
_DATAFRAMEPROTO_TAGSENTRY._options = None
_ATTRIBUTESPROTO_ATTRIBUTESENTRY._options = None
# @@protoc_insertion_point(module_scope)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .exception import PredictException
from .predict_client import PredictClient
from .tf_request import TFRequest
from .tf_request import TFResponse
from .torch_request import TorchRequest
from .torch_request import TorchResponse
from .blade_request import BladeRequest
from .blade_request import BladeResponse 
from .string_request import StringRequest
from .string_request import StringResponse
from .predict_client import ENDPOINT_TYPE_DIRECT
from .predict_client import ENDPOINT_TYPE_VIPSERVER
from .predict_client import ENDPOINT_TYPE_DEFAULT
from .onnx_request import OnnxData
from .onnx_request import OnnxRequest
from .onnx_request import OnnxResponse
from .queue_client import QueueClient
from . import onnx_processor_data_helper as OnnxDataHelper



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import numpy as np
from .onnx_request import OnnxData

TENSOR_TYPE_TO_NP_TYPE = {
    int(OnnxData.TensorProto.FLOAT): np.dtype('float32'),
    int(OnnxData.TensorProto.UINT8): np.dtype('uint8'),
    int(OnnxData.TensorProto.INT8): np.dtype('int8'),
    int(OnnxData.TensorProto.UINT16): np.dtype('uint16'),
    int(OnnxData.TensorProto.INT16): np.dtype('int16'),
    int(OnnxData.TensorProto.INT32): np.dtype('int32'),
    int(OnnxData.TensorProto.INT64): np.dtype('int64'),
    int(OnnxData.TensorProto.DOUBLE): np.dtype('float64'),
    int(OnnxData.TensorProto.UINT32): np.dtype('uint32'),
    int(OnnxData.TensorProto.UINT64): np.dtype('uint64')
}

NP_TYPE_TO_TENSOR_TYPE = {v: k for k, v in TENSOR_TYPE_TO_NP_TYPE.items()}

def to_array(tensor):  # type: (OnnxData.TensorProto) -> np.ndarray[Any]
    """Converts a tensor def object to a numpy array.
    Inputs:
        tensor: a OnnxData.TensorProto object.
    Returns:
        arr: the converted array.
    """
    if tensor.data_type == OnnxData.TensorProto.UNDEFINED:
        raise TypeError("The element type in the input tensor is not defined.")

    tensor_dtype = tensor.data_type
    np_dtype = TENSOR_TYPE_TO_NP_TYPE[tensor_dtype]
    dims = tensor.dims

    if sys.byteorder == 'big':
        # Convert endian from little to big
        convert_endian(tensor)
    # Raw_bytes support: using frombuffer.
    return np.frombuffer(
        tensor.raw_data,
        dtype=np_dtype).reshape(dims)


def from_array(arr, name):  # type: (np.ndarray[Any], Text) -> OnnxData.TensorProto
    """Converts a numpy array to a tensor def.
    Inputs:
        arr: a numpy array.
        name: (optional) the name of the tensor.
    Returns:
        tensor_def: the converted tensor def.
    """
    tensor = OnnxData.TensorProto()
    tensor.dims.extend(arr.shape)
    tensor.name = name

    # For numerical types, directly use numpy raw bytes.
    try:
        dtype = NP_TYPE_TO_TENSOR_TYPE[arr.dtype]
    except KeyError:
        raise RuntimeError(
            "Numpy data type not understood yet: {}".format(str(arr.dtype)))
    tensor.data_type = dtype
    tensor.raw_data = arr.tobytes()  # note: tobytes() is only after 1.9.
    if sys.byteorder == 'big':
        # Convert endian from big to little
        convert_endian(tensor)
    return tensor


def convert_endian(tensor):  # type: (OnnxData.TensorProto) -> None
    """
    call to convert endianess of raw data in tensor.
    Inputs:
        tensor: OnnxData.TensorProto object to be converted.
    """
    tensor_dtype = tensor.data_type
    np_dtype = TENSOR_TYPE_TO_NP_TYPE[tensor_dtype]
    tensor.raw_data = np.frombuffer(tensor.raw_data, dtype=np_dtype).byteswap().tobytes()



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: blade.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='blade.proto',
  package='blade',
  syntax='proto2',
  serialized_options=None,
  serialized_pb=_b('\n\x0b\x62lade.proto\x12\x05\x62lade\"\x18\n\x05Shape\x12\x0f\n\x03\x64im\x18\x01 \x03(\x03\x42\x02\x10\x01\"\xc1\x01\n\x05\x41rray\x12\x17\n\x0f\x62lade_blob_name\x18\x01 \x01(\t\x12\x18\n\x10name_in_dl_model\x18\x02 \x01(\t\x12\x11\n\tbatchsize\x18\x03 \x02(\x05\x12\x1b\n\x05shape\x18\x04 \x01(\x0b\x32\x0c.blade.Shape\x12\x15\n\tfloat_val\x18\x05 \x03(\x02\x42\x02\x10\x01\x12\x13\n\x07int_val\x18\x06 \x03(\x05\x42\x02\x10\x01\x12\x15\n\tint64_val\x18\x07 \x03(\x03\x42\x02\x10\x01\x12\x12\n\nstring_val\x18\x08 \x03(\x0c\"c\n\nOutputInfo\x12\x17\n\x0f\x62lade_blob_name\x18\x01 \x01(\t\x12\x18\n\x10name_in_dl_model\x18\x02 \x01(\t\x12\"\n\tdata_type\x18\x03 \x01(\x0e\x32\x0f.blade.DataType\"j\n\x07Request\x12\x19\n\x11tf_signature_name\x18\x01 \x01(\t\x12\x1c\n\x06inputs\x18\x02 \x03(\x0b\x32\x0c.blade.Array\x12&\n\x0boutput_info\x18\x03 \x03(\x0b\x32\x11.blade.OutputInfo\")\n\x08Response\x12\x1d\n\x07outputs\x18\x01 \x03(\x0b\x32\x0c.blade.Array*S\n\x08\x44\x61taType\x12\x0e\n\nDT_UNKNOWN\x10\x00\x12\x0c\n\x08\x44T_FLOAT\x10\x01\x12\x0c\n\x08\x44T_INT32\x10\x02\x12\x0c\n\x08\x44T_INT64\x10\x03\x12\r\n\tDT_STRING\x10\x04')
)

_DATATYPE = _descriptor.EnumDescriptor(
  name='DataType',
  full_name='blade.DataType',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='DT_UNKNOWN', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DT_FLOAT', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DT_INT32', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DT_INT64', index=3, number=3,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DT_STRING', index=4, number=4,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=496,
  serialized_end=579,
)
_sym_db.RegisterEnumDescriptor(_DATATYPE)

DataType = enum_type_wrapper.EnumTypeWrapper(_DATATYPE)
DT_UNKNOWN = 0
DT_FLOAT = 1
DT_INT32 = 2
DT_INT64 = 3
DT_STRING = 4



_SHAPE = _descriptor.Descriptor(
  name='Shape',
  full_name='blade.Shape',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='dim', full_name='blade.Shape.dim', index=0,
      number=1, type=3, cpp_type=2, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=_b('\020\001'), file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=22,
  serialized_end=46,
)


_ARRAY = _descriptor.Descriptor(
  name='Array',
  full_name='blade.Array',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='blade_blob_name', full_name='blade.Array.blade_blob_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='name_in_dl_model', full_name='blade.Array.name_in_dl_model', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='batchsize', full_name='blade.Array.batchsize', index=2,
      number=3, type=5, cpp_type=1, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='shape', full_name='blade.Array.shape', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='float_val', full_name='blade.Array.float_val', index=4,
      number=5, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=_b('\020\001'), file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='int_val', full_name='blade.Array.int_val', index=5,
      number=6, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=_b('\020\001'), file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='int64_val', full_name='blade.Array.int64_val', index=6,
      number=7, type=3, cpp_type=2, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=_b('\020\001'), file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='string_val', full_name='blade.Array.string_val', index=7,
      number=8, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=49,
  serialized_end=242,
)


_OUTPUTINFO = _descriptor.Descriptor(
  name='OutputInfo',
  full_name='blade.OutputInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='blade_blob_name', full_name='blade.OutputInfo.blade_blob_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='name_in_dl_model', full_name='blade.OutputInfo.name_in_dl_model', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='data_type', full_name='blade.OutputInfo.data_type', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=244,
  serialized_end=343,
)


_REQUEST = _descriptor.Descriptor(
  name='Request',
  full_name='blade.Request',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='tf_signature_name', full_name='blade.Request.tf_signature_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='inputs', full_name='blade.Request.inputs', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='output_info', full_name='blade.Request.output_info', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=345,
  serialized_end=451,
)


_RESPONSE = _descriptor.Descriptor(
  name='Response',
  full_name='blade.Response',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='outputs', full_name='blade.Response.outputs', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=453,
  serialized_end=494,
)

_ARRAY.fields_by_name['shape'].message_type = _SHAPE
_OUTPUTINFO.fields_by_name['data_type'].enum_type = _DATATYPE
_REQUEST.fields_by_name['inputs'].message_type = _ARRAY
_REQUEST.fields_by_name['output_info'].message_type = _OUTPUTINFO
_RESPONSE.fields_by_name['outputs'].message_type = _ARRAY
DESCRIPTOR.message_types_by_name['Shape'] = _SHAPE
DESCRIPTOR.message_types_by_name['Array'] = _ARRAY
DESCRIPTOR.message_types_by_name['OutputInfo'] = _OUTPUTINFO
DESCRIPTOR.message_types_by_name['Request'] = _REQUEST
DESCRIPTOR.message_types_by_name['Response'] = _RESPONSE
DESCRIPTOR.enum_types_by_name['DataType'] = _DATATYPE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Shape = _reflection.GeneratedProtocolMessageType('Shape', (_message.Message,), dict(
  DESCRIPTOR = _SHAPE,
  __module__ = 'blade_pb2'
  # @@protoc_insertion_point(class_scope:blade.Shape)
  ))
_sym_db.RegisterMessage(Shape)

Array = _reflection.GeneratedProtocolMessageType('Array', (_message.Message,), dict(
  DESCRIPTOR = _ARRAY,
  __module__ = 'blade_pb2'
  # @@protoc_insertion_point(class_scope:blade.Array)
  ))
_sym_db.RegisterMessage(Array)

OutputInfo = _reflection.GeneratedProtocolMessageType('OutputInfo', (_message.Message,), dict(
  DESCRIPTOR = _OUTPUTINFO,
  __module__ = 'blade_pb2'
  # @@protoc_insertion_point(class_scope:blade.OutputInfo)
  ))
_sym_db.RegisterMessage(OutputInfo)

Request = _reflection.GeneratedProtocolMessageType('Request', (_message.Message,), dict(
  DESCRIPTOR = _REQUEST,
  __module__ = 'blade_pb2'
  # @@protoc_insertion_point(class_scope:blade.Request)
  ))
_sym_db.RegisterMessage(Request)

Response = _reflection.GeneratedProtocolMessageType('Response', (_message.Message,), dict(
  DESCRIPTOR = _RESPONSE,
  __module__ = 'blade_pb2'
  # @@protoc_insertion_point(class_scope:blade.Response)
  ))
_sym_db.RegisterMessage(Response)


_SHAPE.fields_by_name['dim']._options = None
_ARRAY.fields_by_name['float_val']._options = None
_ARRAY.fields_by_name['int_val']._options = None
_ARRAY.fields_by_name['int64_val']._options = None
# @@protoc_insertion_point(module_scope)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import urllib.parse
import uuid
import json
import logging
import threading
import websocket
import time

from urllib3.exceptions import HTTPError
from urllib3.exceptions import MaxRetryError
from urllib3.exceptions import ProtocolError

from .queue_types.queueservice_pb2 import DataFrameListProto
from .queue_types.queueservice_pb2 import DataFrameProto
from .exception import PredictException
from .predict_client import PredictClient

HeaderRedisUid = "X-EAS-QueueService-Redis-Uid"
HeaderRedisGid = "X-EAS-QueueService-Redis-Gid"
HeaderRequestId = "X-Eas-Queueservice-Request-Id"
HeaderAuthorization = "Authorization"


class DataFrame:
    def __init__(self, df):
        self.raw_data = df
        self.data = df.data
        self.index = df.index
        self.tags = df.tags
        self.message = df.message

    def set_data(self, data):
        self.data = data

    def set_index(self, index):
        self.index = index

    def set_message(self, message):
        self.message = message

    def set_tags(self, tags: dict):
        self.tags = tags

    def __str__(self):
        return self.raw_data.__str__()


class DataFrameList:
    def __init__(self, dataframe_list):
        self.raw_data = dataframe_list
        self.dataframe_list = []

        for i in range(len(dataframe_list.index)):
            df = dataframe_list.index[i]
            self.dataframe_list.append(DataFrame(df))

    def get_list(self):
        return self.dataframe_list

    def __getitem__(self, item):
        return self.dataframe_list[item]

    def __str__(self):
        return self.raw_data.__str__()


class DataFrameCodec:
    # TODO(lingcai.wl): support flatbuffer (maybe not necessary)
    def __init__(self, type="protobuffer"):
        self.type = type
        self.request_data = DataFrameProto()

    @classmethod
    def encode(cls, dataframe: DataFrame):
        request_data = DataFrameProto()
        request_data.data = dataframe.data
        request_data.index = dataframe.index
        request_data.tags = dataframe.tags
        request_data.message = dataframe.message
        return request_data.SerializeToString()

    @classmethod
    def decode(cls, raw_data):
        request_data = DataFrameProto()
        request_data.ParseFromString(raw_data)
        return DataFrame(request_data)

    def to_string(self):
        return self.request_data.SerializeToString()


class DataFrameListCodec:
    # TODO(lingcai.wl): support flatbuffer (maybe not necessary)
    def __init__(self, type="protobuffer"):
        self.type = type
        self.request_data = DataFrameListProto()

    @classmethod
    def encode(cls, dataframe_list: DataFrameList):
        request_data = DataFrameListProto()
        for i, df in enumerate(dataframe_list):
            request_data[i] = DataFrameCodec.encode(df)
        return request_data.SerializeToString()

    @classmethod
    def decode(cls, raw_data):
        request_data = DataFrameListProto()
        request_data.ParseFromString(raw_data)
        return DataFrameList(request_data)

    def to_string(self):
        return self.request_data.SerializeToString()


class Watcher:
    def __init__(self, queue, url, headers, logger):
        self.queue = queue
        self.event = threading.Event()
        self.url = url
        self.headers = headers
        self.logger = logger
        self.closed = False

    def _ping(self):
        while not self.event.wait(2):
            with self.queue.watch_lock:
                if not self.closed:
                    try:
                        self.queue.watch_sock.ping("ping")
                    except Exception as ex:
                        self.logger and self.logger.warning(
                            "send_ping error ({})".format(ex))
                else:
                    self.logger and self.logger.debug(
                        "send_ping routine exited")
                    break

    def run(self):
        with self.queue.watch_lock:
            if self.queue.watch_sock is not None:
                raise PredictException(
                    400, 'Another watcher is already running')
            self.queue.watch_sock = websocket.create_connection(
                self.url, header=self.headers)

        thread = threading.Thread(target=self._ping)
        thread.daemon = True
        thread.start()

        while True:
            try:
                df = self.queue.watch_sock.recv()
                yield DataFrameCodec.decode(df)
            except Exception as ex:
                self.logger and self.logger.warning(
                    "connection closed (%s), try to reconnect" % str(ex))
                while True:
                    try:
                        self.queue.watch_sock = websocket.create_connection(
                            self.url, header=self.headers)
                        break
                    except Exception as e:
                        time.sleep(1)
                        self.logger and self.logger.warning(
                            "reconnect failed (%s), retrying..." % str(e))

    def close(self):
        with self.queue.watch_lock:
            self.event.set()
            self.queue.watch_sock.abort()
            self.queue.watch_sock = None
            self.closed = True


class QueueClient(PredictClient):
    """
    Client for accessing prediction service by creating a fixed size connection pool
    to perform the request through established persistent connections.
    """

    def init(self, uid=None, gid='eas'):
        """
        init async client with uid and group id
        """
        super(QueueClient, self).init()

        if uid is None:
            self.uid = str(uuid.uuid4())
        else:
            self.uid = uid
        self.gid = gid
        self.watch_sock = None
        self.watch_lock = threading.Lock()
        self.logger = logging.getLogger(__name__)
        self.logger.setLevel(logging.INFO)

    def set_logger(self, logger=None):
        self.logger = logger

    def _with_identity(self, headers=None):
        """
        add identity info into headers
        """
        if headers is None:
            headers = {}
        headers[HeaderAuthorization] = self.token
        headers[HeaderRedisUid] = self.uid
        headers[HeaderRedisGid] = self.gid
        return headers

    def _build_url(self, query, websocket=False):
        """
        build the url of target queue service
        """
        domain = self.endpoint.get()
        url = '%s/api/predict/%s' % (domain, self.service_name)

        if websocket:
            url = url.replace('http://', 'ws://').replace('https://', 'wss://')
            if (not url.startswith('ws://')) and (not url.startswith('wss://')):
                url = 'ws://' + url

        query_str = urllib.parse.urlencode(query)
        if len(query_str) != 0:
            url += '?' + query_str
        return url

    def _process_indexes(self, indexes, method):
        """
        common code for indexes related request
        """
        query = {'_indexes_': ','.join([str(ind) for ind in indexes])}
        url = self._build_url(query)
        headers = self._with_identity()
        return self._do_request(url, method, headers).data

    def _do_request(self, url, method, headers, body=None):
        """
        common code for a http request
        """
        try:
            for i in range(self.retry_count + 1):
                resp = self.connection_pool.request(method, url,
                                                    headers=headers,
                                                    body=body,
                                                    timeout=self.timeout / 1000.0,
                                                    retries=self.retry_count)

                # retry for non-200
                if int(resp.status / 100) != 2:
                    continue

                return resp

            raise PredictException(resp.status, resp.data)

        except (MaxRetryError, ProtocolError, HTTPError) as e:
            self.logger and self.logger.debug(
                'Request failed, err: %s', str(e))
            raise PredictException(502, 'url: %s, error: %s' % (url, str(e)))

    def truncate(self, index):
        """
        delete data whose indexes are smaller than the specified index from a queue service
        """
        query = {
            '_index_': index,
            '_trunc_': 'true'
        }
        url = self._build_url(query)

        headers = self._with_identity()
        return self._do_request(url, 'DELETE', headers).data

    def put(self, data, tags: dict = {}):
        """
        put data into a queue service
        """
        url = self._build_url(tags)

        headers = self._with_identity()
        resp = self._do_request(url, 'POST', headers, data)
        request_id = resp.headers[HeaderRequestId]
        return resp.data.decode('utf-8'), request_id

    def attributes(self):
        """
        get the attributes of a queue service
        """
        url = self._build_url({'_attrs_': 'true'})

        headers = self._with_identity()
        resp = self._do_request(url, 'GET', headers)
        return json.loads(resp.data.strip().decode('utf-8'))

    def get(self, request_id=None, index=0, length=1, timeout='5s', auto_delete=True, tags={}):
        """
        get the data in queue service
        """
        query = {
            '_index_': str(index),
            '_length_': str(length),
            '_timeout_': str(timeout),
            '_raw_': 'false',
            '_auto_delete_': str(auto_delete).lower(),
        }
        if request_id is not None:
            query['requestId'] = request_id

        query.update(tags)

        url = self._build_url(query)

        headers = self._with_identity()
        headers['Accept'] = 'application/vnd.google.protobuf'  # default

        resp = self._do_request(url, 'GET', headers)
        dfl = DataFrameListCodec.decode(resp.data)
        return dfl.get_list()

    def delete(self, index):
        """
        delete indexes
        """
        if not isinstance(index, list):
            index = [index]
        return self._process_indexes(index, 'DELETE').decode('utf-8')

    def watch(self, index: int, window: int, index_only=False, auto_commit=False, tags=None):
        """
        create a watcher to read data items from a queue
        """
        if tags is None:
            tags = {}

        query = {
            '_watch_': 'true',
            '_index_': index,
            '_window_': window,
            '_index_only_': str(index_only).lower(),
            '_auto_commit_': str(auto_commit).lower(),
        }
        query.update(tags)
        url = self._build_url(query, websocket=True)

        headers = self._with_identity()
        headers['Accept'] = 'application/vnd.google.protobuf'  # default

        try:
            return Watcher(self, url, headers, self.logger)
        except websocket._exceptions.WebSocketException as e:
            self.watch_sock = None
            raise PredictException(400, 'url: %s, error: %s' % (url, str(e)))

    def commit(self, index):
        """
        commit indexes
        """
        if not isinstance(index, list):
            index = [index]
        return self._process_indexes(index, 'PUT')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .request import Request
from .request import Response
from . import tf_request_pb2 as tf_pb


class TFRequest(Request):
    """
    Request for tensorflow services whose input data is in format of protobuf,
    privide methods to generate the required protobuf object, and serialze it to string
    """
    DT_FLOAT = tf_pb.DT_FLOAT
    DT_DOUBLE = tf_pb.DT_DOUBLE
    DT_INT8 = tf_pb.DT_INT8
    DT_INT16 = tf_pb.DT_INT16
    DT_INT32 = tf_pb.DT_INT32
    DT_INT64 = tf_pb.DT_INT64
    DT_UINT8 = tf_pb.DT_UINT8
    DT_UINT16 = tf_pb.DT_UINT16
    DT_QINT8 = tf_pb.DT_QINT8
    DT_QUINT8 = tf_pb.DT_QUINT8
    DT_QINT16 = tf_pb.DT_QINT16
    DT_QUINT16 = tf_pb.DT_QUINT16
    DT_QINT32 = tf_pb.DT_QINT32
    DT_STRING = tf_pb.DT_STRING
    DT_BOOL = tf_pb.DT_BOOL

    def __init__(self, signature_name=None):
        self.request_data = tf_pb.PredictRequest()
        self.signature_name = signature_name

    def __str__(self):
        return self.request_data

    def set_signature_name(self, singature_name):
        """
        Set the signature name of the model
        :param singature_name: signature name of the model
        """
        self.signature_name = singature_name

    def add_feed(self, input_name, shape, content_type, content):
        """
        Add input data for the request, a tensorflow model may have many inputs with different
        data types, this methods set data for one of the input with the specified name 'input_name'
        :param input_name: name of the input to be set
        :param shape: shape of the input tensor in format of array, such as [1,784]
        :param content_type: type of the input tensor, can be one of the predefined data type, such as TFRequest.DT_FLOAT
        :param content: data content of the input tensor, which is expanded to one-dimension array, such as [1,2,3,4,5]
        """
        self.request_data.signature_name = self.signature_name
        self.request_data.inputs[input_name].dtype = content_type
        self.request_data.inputs[input_name].array_shape.dim.extend(shape)
        integer_types = [
            tf_pb.DT_INT8,
            tf_pb.DT_INT16,
            tf_pb.DT_INT32 ,
            tf_pb.DT_UINT8 ,
            tf_pb.DT_UINT16,
            tf_pb.DT_QINT8,
            tf_pb.DT_QINT16,
            tf_pb.DT_QINT32,
            tf_pb.DT_QUINT8,
            tf_pb.DT_QUINT16,
        ]
        if content_type == tf_pb.DT_FLOAT:
            self.request_data.inputs[input_name].float_val.extend(content)
        elif content_type == tf_pb.DT_DOUBLE:
            self.request_data.inputs[input_name].double_val.extend(content)
        elif content_type in integer_types:
            self.request_data.inputs[input_name].int_val.extend(content)
        elif content_type == tf_pb.DT_INT64:
            self.request_data.inputs[input_name].int64_val.extend(content)
        elif content_type == tf_pb.DT_BOOL:
            self.request_data.inputs[input_name].bool_val.extend(content)
        elif content_type == tf_pb.DT_STRING:
            self.request_data.inputs[input_name].string_val.extend(content)

    def add_fetch(self, output_name):
        """
        Add output node name for the request to get, if not specified, then all the known outputs are fetched,
        but for frozen models, the output name must be specified, or else the service would throw exception like:
        'Must specify at least one target to fetch or execute.'
        :param output_name: name of the output node to fetch
        """
        self.request_data.output_filter.append(output_name)

    def to_string(self):
        """
        Serialize the request to string for transmission
        :return: the request data in format of string
        """
        return self.request_data.SerializeToString()

    def parse_response(self, response_data):
        """
        Parse the given response data in string format to the related TFResponse object
        :param response_data: the service response data in string format
        :return: the TFResponse object related the request
        """
        return TFResponse(response_data)


class TFResponse(Response):
    """
    Deserialize the response data to a structured object for users to read
    """

    def __init__(self, response_data=None):
        self.response = tf_pb.PredictResponse()
        self.response.ParseFromString(response_data)

    def __str__(self):
        return str(self.response)

    def get_tensor_shape(self, output_name):
        """
        Get the shape of a specified output tensor
        :param output_name: name of the output tensor
        :return: the shape in format of list
        """
        return list(self.response.outputs[output_name].array_shape.dim)

    def get_values(self, output_name):
        """
        Get the value of a specified output tensor
        :param output_name: name of the output tensor
        :return: the content of the output tensor
        """
        output = self.response.outputs[output_name]
        if output.dtype == TFRequest.DT_FLOAT:
            return output.float_val
        elif output.dtype == TFRequest.DT_INT8 or output.dtype == TFRequest.DT_INT16 or \
                output.dtype == TFRequest.DT_INT32:
            return output.int_val
        elif output.dtype == TFRequest.DT_INT64:
            return output.int64_val
        elif output.dtype == TFRequest.DT_DOUBLE:
            return output.double_val
        elif output.dtype == TFRequest.DT_STRING:
            return output.string_val
        elif output.dtype == TFRequest.DT_BOOL:
            return output.bool_val



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .request import Request
from .request import Response
from . import onnx_processor_data_pb2 as OnnxData


class OnnxRequest(Request):
    """
    Request for hie onnx processor services whose input data is in format of protobuf,
    privide methods to generate the required protobuf object, and serialze it to string
    """

    def __init__(self):
        self.request_data = OnnxData.DataProto()
        self.name_index_map = {}

    def __str__(self):
        return str(self.request_data)

    def add_tensor(self, tensor): # type: (OnnxData.TensorProto) -> int
        """
        Add input tensor for the request
        :param tensor: a OnnxData.TensorProto object
        :return: this tensor's index in request
        """
        name = tensor.name
        if name in self.name_index_map:
            index = self.name_index_map[name]
        else:
            index = len(self.request_data.tensors)
            self.name_index_map[name] = index
            self.request_data.tensors.add()
        self.request_data.tensors[index].CopyFrom(tensor)
        return index

    def to_string(self):
        """
        Serialize the request to string for transmission
        :return: the request data in format of string
        """
        return self.request_data.SerializeToString()

    def parse_response(self, response_data):
        """
        Parse the given response data in string format to the related OnnxResponse object
        :param response_data: the service response data in string format
        :return: the OnnxResponse object related the request
        """
        return OnnxResponse(response_data)


class OnnxResponse(Response):
    """
    Deserialize the response data to a structured object for users to read
    """

    def __init__(self, response_data=None):
        self.response = OnnxData.DataProto()
        self.response.ParseFromString(response_data)

    def __str__(self):
        return str(self.response)

    def get_size(self):
        """
        Get the number of tensors of a response
        :return: the tensors size in format of int
        """
        return len(self.response.tensors)

    def get_name(self, output_index): # type: (int) -> OnnxData.TensorProto
        """
        Get the name of a specified output tensor
        :param output_index: name of the output tensor
        :return: the name in format of string
        """
        return self.response.tensors[output_index].name

    def get_tensor(self, output_index): # type: (int) -> OnnxData.TensorProto
        """
        Get the shape of a specified output tensor
        :param output_index: name of the output tensor
        :return: the shape in format of list
        """
        return self.response.tensors[output_index]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .request import Request
from .request import Response
from . import blade_pb2
from functools import reduce 
import operator


class BladeRequest(Request):
    """
    Request for tensorflow services whose input data is in format of protobuf,
    privide methods to generate the required protobuf object, and serialze it to string
    """
    DT_FLOAT = blade_pb2.DT_FLOAT
    DT_INT32 = blade_pb2.DT_INT32
    DT_INT64 = blade_pb2.DT_INT64
    DT_STRING = blade_pb2.DT_STRING

    def __init__(self, signature_name=None):
        self.request_data = blade_pb2.Request()
        self.set_signature_name(signature_name)

    def __str__(self):
        return self.request_data

    def set_signature_name(self, signature_name):
        """
        Set the signature name of the model
        :param signature_name: signature name of the tensorflow model
        """
        self.tf_signature_name = signature_name
        self.request_data.tf_signature_name = self.tf_signature_name

    def volume(self, shape):
        return reduce(operator.mul, shape)

    def add_feed(self, input_name, batchsize, shape, content_type, content):
        """
        Add input data for the request, a tensorflow model may have many inputs with different
        data types, this methods set data for one of the input with the specified name 'input_name'
        :param input_name: name of the input to be set
        :param batchsize: batch size of the input tensor
        :param shape: shape of the input tensor in format of array, such as [1,784]
        :param content_type: type of the input tensor (predefined data type, such as BladeRequest.DT_FLOAT)
        :param content: data content of the input tensor, the content could be a single value (e.g. 3.14), or a expanded one-dimension array (e.g. [1,2,3,4,5])
        """
        if (not input_name):
            raise AssertionError('"input name" can not be emtpy')
        if (not isinstance(batchsize, int) or batchsize <= 0):
            raise AssertionError('Invalid batchsize value: {}'.format(batchsize))
        if (not isinstance(shape, list)):
            raise AssertionError('Invalid type for "shape": {}'.format(type(shape)))
        arr = self.request_data.inputs.add()
        arr.shape.dim.extend(shape)
        arr.name_in_dl_model = input_name
        arr.blade_blob_name = input_name # HERE: 'blade blob name' is identical to 'name in DL model'.
        arr.batchsize = batchsize
        if (content_type == blade_pb2.DT_FLOAT):
            if (isinstance(content, float)):
                if (int(self.volume(shape)) == 1):
                    arr.float_val.append(content)
            elif (isinstance(content, list)):
                if (int(self.volume(shape)) == int(len(content))):
                    arr.float_val.extend(content)
                else:
                    raise AssertionError('Shape volume {} mis-match content size {}'.format( \
                            self.volume(shape), len(content)))
            else:
                raise AssertionError('Expect float or list of float type')
        elif (content_type == blade_pb2.DT_INT32):
            if (isinstance(content, int)):
                if (int(self.volume(shape)) == 1):
                    arr.int_val.append(content)
                else:
                    raise AssertionError('Shape volume {} mis-match content size {}'.format( \
                            self.volume(shape), 1))
            elif (isinstance(content, list)):
                if (int(self.volume(shape)) == int(len(content))):
                    arr.int_val.extend(content)
                else:
                    raise AssertionError('Shape volume {} mis-match content size {}'.format( \
                            self.volume(shape), len(content)))
            else:
                raise AssertionError('Expect int32 or list of int32 type')
        elif (content_type == blade_pb2.DT_INT64):
            if (isinstance(content, int)):
                if (int(self.volume(shape)) == 1):
                    arr.in64_val.append(content)
                else:
                    raise AssertionError('Shape volume {} mis-match content size {}'.format( \
                            self.volume(shape), 1))
            elif (isinstance(content, list)):
                if (int(self.volume(shape)) == int(len(content))):
                    arr.int64_val.extend(content)
                else:
                    raise AssertionError('Shape volume {} mis-match content size {}'.format( \
                            self.volume(shape), len(content)))
            else:
                raise AssertionError('Expect int64 or list of int64 type')
        elif (content_type == blade_pb2.DT_STRING):
            if (isinstance(content, str)):
                if (int(self.volume(shape)) == 1):
                    arr.string_val.append(content)
                else:
                    raise AssertionError('Shape volume {} mis-match content size {}'.format( \
                            self.volume(shape), 1))
            elif (isinstance(content, list)):
                if (int(self.volume(shape)) == int(len(content))):
                    arr.string_val.extend(content)
                else:
                    raise AssertionError('Shape volume {} mis-match content size {}'.format( \
                            self.volume(shape), len(content)))
            else:
                raise AssertionError('Expect str or list of str type')
        else:
            raise AssertionError('Unknown data_type: "{}"'.format(data_type))
        return

    def add_fetch(self, output_name, output_type):
        """
        Add output node name for the request to get, if not specified, then all the known outputs are fetched,
        but for frozen models, the output name must be specified, or else the service would throw exception like:
        'Must specify at least one target to fetch or execute.'
        :param output_name: name of the output node to fetch
        :param output_type: type of the output tensor (predefined data type, e.g BladeRequest.DT_FLOAT)
        """
        if (not output_name):
            raise AssertionError('"input name" can not be emtpy')
        if (output_type not in blade_pb2.DataType.values()):
            raise AssertionError('Unknown data_type: "{}"'.format(data_type))
        out_info = self.request_data.output_info.add()
        out_info.blade_blob_name = output_name
        out_info.name_in_dl_model = output_name
        out_info.data_type = output_type
        return

    def to_string(self):
        """
        Serialize the request to string for transmission
        :return: the request data in format of string
        """
        return self.request_data.SerializeToString()

    def parse_response(self, response_data):
        """
        Parse the given response data in string format to the related BladeResponse object
        :param response_data: the service response data in string format
        :return: the BladeResponse object related the request
        """
        return BladeResponse(response_data)


class BladeResponse(Response):
    """
    Deserialize the response data to a structured object for users to read
    """

    def __init__(self, response_data=None):
        self.response = blade_pb2.Response()
        self.response.ParseFromString(response_data)

    def __str__(self):
        return str(self.response)

    def get_tensor_shape(self, output_name):
        """
        Get the shape of a specified output tensor
        :param output_name: name of the output tensor
        :return: the shape in format of list
        """
        for out in self.response.outputs:
            if (str(out.name_in_dl_model).strip() == str(output_name).strip()):
                return list(out.shape.dim)
        raise AssertionError('Unknown output name: "{}"'.format(output_name))

    def get_values(self, output_name, output_type):
        """
        Get the value of a specified output tensor
        :param output_name: name of the output tensor
        :param output_type: type of the output tensor (predefined data type, e.g BladeRequest.DT_FLOAT)
        :return: the content of the output tensor
        """
        if (output_type not in blade_pb2.DataType.values()):
            raise AssertionError('Unknown data_type: "{}"'.format(data_type))
        for out in self.response.outputs:
            if (str(out.name_in_dl_model).strip() == str(output_name).strip()):
                if (output_type == BladeRequest.DT_FLOAT):
                    return out.float_val
                elif (output_type == BladeRequest.DT_INT32):
                    return out.int_val
                elif (output_type == BladeRequest.DT_INT64):
                    return out.int64_val
                elif (output_type == BladeRequest.DT_STRING):
                    return out.string_val
        raise AssertionError('Unknown output name: "{}"'.format(output_name))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import threading
import time
import sys
import logging
import datetime
import hashlib
import hmac
import base64
from urllib3 import PoolManager
from urllib3.exceptions import MaxRetryError
from urllib3.exceptions import ProtocolError
from urllib3.exceptions import HTTPError
from .exception import PredictException
from .vipserver_endpoint import VipServerEndpoint
from .cacheserver_endpoint import CacheServerEndpoint
from .gateway_endpoint import GatewayEndpoint

ENDPOINT_TYPE_DIRECT = 'DIRECT'
ENDPOINT_TYPE_VIPSERVER = 'VIPSERVER'
ENDPOINT_TYPE_DEFAULT = 'DEFAULT'


class PredictClient:
    """
    Client for accessing prediction service by creating a fixed size connection pool
    to perform the request through established persistent connections.
    """

    def __init__(self, endpoint='', service_name=''):
        self.retry_count = 5
        self.max_connection_count = 100
        self.token = ''
        self.endpoint = None
        self.timeout = 5000
        self.connection_pool = None
        self.endpoint_type = ''
        self.endpoint_name = endpoint
        self.service_name = service_name
        self.stop = False
        self.logger = logging.getLogger(endpoint + '/' + service_name)
        self.logger.addHandler(logging.StreamHandler(stream=sys.stdout))
        self.logger.setLevel(logging.ERROR)

    def __sync_handler(self):
        while True:
            if self.stop:
                break
            try:
                self.endpoint.sync()
            except PredictException as e:
                self.logger.error(str(e))
            time.sleep(3)

    def destroy(self):
        self.stop = True

    def init(self):
        """
        Initialize the client after the functions used to set the client properties are called
        """
        if self.connection_pool is None:
            self.connection_pool = PoolManager(maxsize=self.max_connection_count)

        if self.endpoint_type == '' or self.endpoint_type == ENDPOINT_TYPE_DEFAULT:
            self.endpoint = GatewayEndpoint(self.endpoint_name, self.service_name, self.logger)
        elif self.endpoint_type == ENDPOINT_TYPE_VIPSERVER:
            self.endpoint = VipServerEndpoint(self.endpoint_name, self.logger)
        elif self.endpoint_type == ENDPOINT_TYPE_DIRECT:
            self.endpoint = CacheServerEndpoint(self.endpoint_name, self.service_name, self.logger)
        else:
            raise PredictException(500, 'Unsupported endpoint type: %s' % self.endpoint_type)

        t = threading.Thread(target=PredictClient.__sync_handler, args=(self,))
        t.daemon = True
        t.start()
        self.logger.debug('Endpoint sync thread started')

    def set_endpoint(self, endpoint):
        """
        Set the endpoint of the service for the client
        :param endpoint_name: name of the endpoint, such as http://pai-eas-vpc.cn-shanghai.aliyuncs.com
        """
        self.endpoint_name = endpoint

    def set_service_name(self, service_name):
        """
        Set the service name for the client
        :param service_name: name of the service to access
        """
        self.service_name = service_name

    def set_endpoint_type(self, endpoint_type):
        """
        Set the endpoint type, support GATEWAY, DIRECT, VIPSERVER, default is GATEWAY
        :param endpoint_type: type of the endpoint
        """
        self.endpoint_type = endpoint_type

    def set_token(self, token):
        """
        Set the authentication token of the service for the client
        :param token: service token, automatically generated when deploying the service
        """
        self.token = token

    def set_log_level(self, log_level):
        """
        Set log level for logging module
        :param log_level: target log level
        """
        self.logger.setLevel(log_level)

    def set_retry_count(self, count):
        """
        Set the max count of retrying when an error occurred during a request
        :param count: max retry count
        """
        self.retry_count = count

    def set_max_connection_count(self, count):
        """
        Set the max connection count which is the upper limit of connections in the connection pool
        :param count: max connection count
        """
        self.max_connection_count = count

    def set_timeout(self, timeout):
        """
        Set the request timeout for the client
        :param timeout: timeout of a single request
        :return:
        """
        self.timeout = timeout

    def generate_singaure(self, request_data):
        canonicalized_resource = '/api/predict/%s' % self.service_name

        content_md5 = hashlib.md5(request_data).hexdigest()
        content_type = 'application/octet-stream'
        utcnow = datetime.datetime.now()
        current_time = utcnow.strftime('%a, %d %b %Y %H:%M:%S GMT')
        verb = 'POST'

        auth = '%s\n%s\n%s\n%s\n%s' % (verb, content_md5, content_type, current_time, canonicalized_resource)

        hmac1 = hmac.new(bytearray(self.token, 'utf-8'), bytearray(auth.strip(), 'utf-8'), hashlib.sha1)
        signature = base64.b64encode(hmac1.digest()).decode()
        authorization = 'EAS %s' % (signature.strip())

        return {
            'Content-MD5': content_md5,
            'Date': current_time,
            'Content-Type': content_type,
            'Content-Length': '%d' % len(request_data),
            'Authorization': authorization
        }

    def predict(self, req):
        """
        Perform the prediction request to the server by sending an http request of which the request body
        may be in different format (string, protobuf or other user defined format), make it an abstract
        class implementing a basic function 'to_string()' to serialized the request to string for sending
        :param req: abstract class of the request
        :return: service response correlated with the input request
        """
        headers = None
        for i in range(0, self.retry_count):
            try:
                domain = self.endpoint.get()
                url = u'%s/api/predict/%s' % (domain, self.service_name)
                self.logger.debug('Request to url: %s' % url)
                req_str = req.to_string()
                if sys.version_info[0] == 3 and isinstance(req_str, str):
                    req_body = bytearray(req_str, 'utf-8')
                else:
                    req_body = bytearray(req_str)
                if len(self.token) > 0:
                    headers = self.generate_singaure(req_body)
                resp = self.connection_pool.request('POST', url,
                                                    headers=headers,
                                                    body=req_body,
                                                    timeout=self.timeout / 1000.0,
                                                    retries=0)
                if resp.status / 100 == 5:
                    if i != self.retry_count - 1:
                        continue
                    raise PredictException(resp.status, resp.data)

                if resp.status != 200:
                    raise PredictException(resp.status, resp.data)

                return req.parse_response(resp.data)
            except (MaxRetryError, ProtocolError, HTTPError) as e:
                self.logger.debug('Request failed, err: %s, retrying', str(e))
                if i != self.retry_count - 1:
                    continue
                raise PredictException(500, 'url: %s, error: %s' % (url, str(e)))



#!/usr/bin/env python
# -*- coding: utf-8 -*-


class PredictException(Exception):
    def __init__(self, code, message):
        self.code = code
        self.message = message

    def __str__(self):
        return 'Code: %s, Message: %s' % (self.code, self.message)


#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .exception import PredictException
from .weighted_round_robin import WRRScheduler
from threading import Lock
import time
import traceback


def endpoint_sort_func(x):
    return '%s:%s:%s' % (x[0]['ip'], x[0]['port'], x[1])

class Endpoint(object):
    r_lock = Lock()
    w_lock = Lock()
    num_r = 0

    def __init__(self, logger):
        self.endpoints = dict()
        self.scheduler = None
        self.logger = logger

    def __r_lock_acquire(self):
        self.r_lock.acquire()
        self.num_r += 1
        if self.num_r == 1:
            self.w_lock.acquire()
        self.r_lock.release()

    def __r_lock_release(self):
        self.r_lock.acquire()
        self.num_r -= 1
        if self.num_r == 0:
            self.w_lock.release()
        self.r_lock.release()

    def __w_lock_acquire(self):
        self.w_lock.acquire()

    def __w_lock_release(self):
        self.w_lock.release()

    def __changed(self, endpoints):
        if len(self.endpoints) != len(endpoints):
            return True
        endpoints = sorted(endpoints, key=endpoint_sort_func)
        for idx, ep in enumerate(self.endpoints):
            old_endpoint = ep[0]
            old_weight = ep[1]
            cur_endpoint = endpoints[idx][0]
            cur_weight = endpoints[idx][1]
            if old_weight != cur_weight:
                return True
            if old_endpoint['ip'] != cur_endpoint['ip'] or old_endpoint['port'] != cur_endpoint['port']:
                return True
        return False

    def set_endpoints(self, endpoints):
        if self.scheduler is None:
            self.__w_lock_acquire()

            try:
                self.scheduler = WRRScheduler(endpoints)
                self.endpoints = sorted(endpoints, key=endpoint_sort_func)
            except Exception as e:
                traceback.print_stack()

            self.__w_lock_release()
            self.logger.info('Service endpoints initialized to: %s' % endpoints)
        elif self.__changed(endpoints):
            self.__w_lock_acquire()

            try:
                self.scheduler.set_data(endpoints)
                self.endpoints = sorted(endpoints, key=endpoint_sort_func)
            except Exception as e:
                traceback.print_stack()

            self.__w_lock_release()
            self.logger.info('Service endpoints changed to: %s' % endpoints)
        else:
            self.logger.debug('Service endpoints unchanged, skip it')

    def get_size(self):
        self.__r_lock_acquire()
        length = len(self.endpoints)
        self.__r_lock_release()
        return length

    def sync(self):
        pass

    def get(self):
        retry = 0
        while True:
            if self.scheduler is not None:
                break
            time.sleep(0.1)
            retry += 1
            if retry > 50: # 5s timeout
                raise PredictException(500, 'Get service backend timeout')
                break

        self.__w_lock_acquire()
        try:
            ep = self.scheduler.get_next()
            if ep is not None:
                ep = ep[0]
        # catch every unexpected exceptions in case of dead lock
        except Exception as e:
            ep = None
            traceback.print_stack()
            self.logger.error('Unexpected exception occured while getting item from WRR: %s' % str(e))
        self.__w_lock_release()

        if ep is None:
            raise PredictException(500, 'No backend found for the target service')

        return ep['ip'] + ':' + str(ep['port'])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest

from .predict_client import PredictClient
from .predict_client import ENDPOINT_TYPE_VIPSERVER
from .predict_client import ENDPOINT_TYPE_DIRECT
from .string_request import StringRequest
from .tf_request import TFRequest
from .torch_request import TorchRequest
from .exception import PredictException

import os
import logging
logging.basicConfig(level = logging.ERROR,format = '%(asctime)s - %(name)s - %(levelname)s - %(message)s')
logger = logging.getLogger(__name__)


class PredictClientTestCase(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super(PredictClientTestCase, self).__init__(*args, **kwargs)
        self.client = PredictClient()

    def setUp(self):
        self.client.set_endpoint('http://pai-eas-vpc.cn-shanghai.aliyuncs.com')
        self.client.set_service_name('scorecard_pmml_example')
        self.client.set_token('YWFlMDYyZDNmNTc3M2I3MzMwYmY0MmYwM2Y2MTYxMTY4NzBkNzdjOQ==')
        self.client.init()

    def tearDown(self):
        self.client.destroy()

    def test_predict_pmml_empty(self):
        req = StringRequest()
        ex = None
        try:
            self.client.predict(req)
        except PredictException as ex:
            pass

        self.assertEqual(ex, None)

    def test_prediction_pmml_bad_json(self):
        req = StringRequest('[{]]')
        ex = None
        try:
            self.client.predict(req)
        except Exception as e:
            ex = e

        self.assertEqual(True, isinstance(ex, PredictException))

    def test_prediction_gateway_no_endpoint(self):
        ex = None
        try:
            client = PredictClient('', 'echo_vipserver')
            client.init()
        except Exception as e:
            ex = e

        self.assertEqual(True, isinstance(ex, PredictException))

    def test_prediction_gateway_endpoint(self):
        os.environ['NAMESPACE'] = 'echo-test'
        client = PredictClient('', 'echo_test')
        client.init()
        endpoint = client.endpoint.get()
        self.assertEqual('http://echo-test.echo-test:8080', endpoint)
        del os.environ['NAMESPACE']

    def test_prediction_bad_endpoint(self):
        req = StringRequest('[{}]')
        response = self.client.predict(req)


#    def test_prediction_direct(self):
#        self.client.set_endpoint_type(ENDPOINT_TYPE_DIRECT)
#        self.client.init()
#
#        request = StringRequest('[{}]')
#        response = self.client.predict(request)


    def test_prediction_tensorflow(self):
        self.client.set_service_name('mnist_saved_model_example')
        self.client.set_token('YTg2ZjE0ZjM4ZmE3OTc0NzYxZDMyNmYzMTJjZTQ1YmU0N2FjMTAyMA==')
        self.client.init()

        req = TFRequest('predict_images')
        req.add_feed('images', [1, 784], TFRequest.DT_FLOAT, [8] * 784)
        response = self.client.predict(req)


    def test_prediction_vipserver(self):
        client = PredictClient('echo-vipserver.shanghai.eas.vipserver', 'echo_vipserver')
        client.set_endpoint_type(ENDPOINT_TYPE_VIPSERVER)
        client.set_token('MDAwZDQ3NjE3OThhOTI4ODFmMjJiYzE0MDk1NWRkOGI1MmVhMGI0Yw==')
        client.init()

        request = StringRequest('[{}]')
        response = self.client.predict(request)


    def test_vipserver_endpoint(self):
        from .vipserver_endpoint import VipServerEndpoint
        ep = VipServerEndpoint('echo-vipserver.shanghai.eas.vipserver', logger)

    def test_vipserver_sync(self):
        from .vipserver_endpoint import VipServerEndpoint
        endpoint = VipServerEndpoint('echo-vipserver.shanghai.eas.vipserver', logger)
        endpoint.sync()
        endpoint.get()


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .endpoint import Endpoint
import urllib3
import json
import os


class CacheServerEndpoint(Endpoint):
    def __init__(self, domain, service_name, logger):
        super(CacheServerEndpoint, self).__init__(logger)
        self.domain = domain
        self.service_name = service_name.split('/')[0]
        self.http = urllib3.PoolManager()
        self.logger = logger
        self.logger.info('Service discovery endpoint is: %s' % self.domain)

    def sync(self):
        self.domain = self.domain.replace('http://', '')
        self.domain = self.domain.replace('https://', '')
        namespace = os.getenv('NAMESPACE')
        pod_name = os.getenv('POD_NAME')
        internal = False
        if namespace is not None and pod_name is not None:
            internal = True
        url = 'http://%s/exported/apis/eas.alibaba-inc.k8s.io/v1/upstreams/%s%s' % \
              (self.domain, self.service_name, '?internal=true' if internal else '')
        endpoints = []
        try:
            resp = self.http.request('GET', url)
            if resp.status != 200:
                self.logger.error('sync service endpoints error: %s, %s' % (resp.status, resp.data))
                return

            resp_data = resp.data.decode('utf-8')
            result = json.loads(resp_data)
            hosts = result['endpoints']['items']
            for host in hosts:
                endpoints.append(({
                                      'ip': host['ip'],
                                      'port': host['port'],
                                  }, host['weight']))
            self.logger.debug(endpoints)
            self.set_endpoints(endpoints)
        except urllib3.exceptions.HTTPError as e:
            self.logger.error('sync service endpoints http error, [%s]: %s' % (url, str(e)))
        except Exception as e:
            self.logger.error('sync service endpoints error, [%s]: %s' % (url, str(e)))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup

setup(
    name='eas-prediction',
    version='0.22',
    description='Python client sdk for eas inference service',
    url='http://gitlab.alibaba-inc.com/eas/eas-python-sdk',
    packages=['eas_prediction', 'eas_prediction/queue_types'],
    install_requires=['urllib3', 'protobuf>=3.14.0, <=3.20.3', 'numpy>=1.9', 'websocket-client'])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from eas_prediction.request import Request
from eas_prediction.request import Response
from eas_prediction.easyrec_predict_pb2 import PBRequest, PBResponse

class EasyRecRequest(Request):
    """
    Request for tensorflow services whose input data is in format of protobuf,
    privide methods to generate the required protobuf object, and serialze it to string
    """

    def __init__(self, signature_name=None, fg_mode="normal"):
        self.fg_mode = fg_mode
        self.request_data = PBRequest()
        self.signature_name = signature_name

    def __str__(self):
        return self.request_data

    def set_signature_name(self, singature_name):
        """
        Set the signature name of the model
        :param singature_name: signature name of the model
        """
        self.signature_name = singature_name

    def add_feed(self, data, debug_level=0):
        if isinstance(data, PBRequest):       
          self.request_data = data
        else:
          self.request_data.ParseFromString(data)
        self.request_data.debug_level = debug_level

    def add_user_fea_flt(self, k, v):
        self.request_data.user_features[k].float_feature = float(v)

    def add_user_fea_int(self, k, v):
        self.request_data.user_features[k].int_feature = int(v)

    def add_user_fea_str(self, k, v):
        self.request_data.user_features[k].string_feature = str(v)

    def set_faiss_neigh_num(self, neigh_num):
        self.request_data.faiss_neigh_num = neigh_num

    def keep_one_item_ids(self):
        item_id = self.request_data.item_ids[0]
        self.request_data.ClearField('item_ids')
        self.request_data.item_ids.extend([item_id])

    def to_string(self):
        """
        Serialize the request to string for transmission
        :return: the request data in format of string
        """
        return self.request_data.SerializeToString()

    def parse_response(self, response_data):
        """
        Parse the given response data in string format to the related TFResponse object
        :param response_data: the service response data in string format
        :return: the TFResponse object related the request
        """
        self.response = PBResponse()
        self.response.ParseFromString(response_data)
        return self.response




#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .endpoint import Endpoint
from .exception import PredictException
import urllib3
import random
import json
import re


class VipServerEndpoint(Endpoint):
    def __init__(self, domain, logger):
        super(VipServerEndpoint, self).__init__(logger)
        self.domain = domain
        self.http = urllib3.PoolManager()
        self.logger = logger

    def get_server(self):
        vipserver_endpoint = 'http://jmenv.tbsite.net:8080/vipserver/serverlist'
        try:
            resp = self.http.request('GET', vipserver_endpoint)
            if resp.status != 200:
                raise PredictException(resp.status, resp.data)
            servers = str(resp.data)
            server_list = re.split('([0-9]+\.[0-9]+\.[0-9]+\.[0-9]+)', servers)
            server_list = [x for x in server_list if re.match('[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+', x) is not None ]
        except Exception as e:
            raise PredictException(500, str(e))
        return server_list[random.randint(0, len(server_list) - 1)]

    def sync(self):
        server = self.get_server()
        url = 'http://%s/vipserver/api/srvIPXT?dom=%s&clusters=DEFAULT' % (server, self.domain)
        endpoints = []
        try:
            resp = self.http.request('GET', url)
            if resp.status != 200:
                self.logger.error('sync service endpoints error: %s, %s' % (resp.status, resp.data))
                return

            if isinstance(resp.data, bytes):
                result = json.loads(resp.data.decode('utf-8'))
            else:
                result = json.loads(resp.data)
            hosts = result['hosts']
            for host in hosts:
                if host['valid']:
                    endpoints.append(({
                        'ip': host['ip'],
                        'port': host['port'],
                    }, host['weight']))
            self.logger.debug(endpoints)
            self.set_endpoints(endpoints)
        except urllib3.exceptions.HTTPError as e:
            self.logger.error('sync service endpoints http error, [%s]: %s' % (url, str(e)))
        except Exception as e:
            self.logger.error('sync service endpoints unknown error, [%s]: %s' % (url, str(e)))
